// *********************************************************
//
// acousticApp1.cc : Acoustic App for multiple queries case
// author          : Fabio Silva
//
// $Id: acousticApp1.cc,v 1.1 2001/11/21 19:51:25 fabio Exp $
//
// *********************************************************

#include "nested.hh"
#include <unistd.h>

extern char *application_id;

AcousticApp *acoustic_app;

int MyTimerReceive::recv(handle hdl, void *p)
{
  return acoustic_app->ProcessTimers(hdl, p);
}

void MyTimerReceive::del(void *p)
{
  // Nothing to delete !
}

void MyReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  acoustic_app->recv(data, my_handle);
}

void AcousticApp::recv(NRAttrVec *data, handle h)
{
  if (h == userSubscribeHandle){
    ProcessUserRequest(data);
    return;
  }

  diffPrint(DEBUG_ALWAYS, "Received a message for an unknown handle !\n");
}

int AcousticApp::ProcessTimers(handle hdl, void *p)
{
  TimerType *timer;

  timer = (TimerType *) p;

  switch (timer->which_timer){

  case STOP_TIMER:

    timetoStop();
    exit(0);

    break;
  }
  return 0;
}

void AcousticApp::timetoStop()
{
  struct timeval tv;

  getTime(&tv);
  fprintf(ofile, "%ld:%ld - Acoustic Node %d stopping\n", tv.tv_sec, tv.tv_usec, node_id);
  fprintf(ofile, "Total messages sent : %d\n", send_counter);
  fflush(NULL);
  fclose(ofile);
}

void AcousticApp::ProcessUserRequest(NRAttrVec *data)
{
  struct timeval tmv;
  NRAttrVec *attrs;
  NRSimpleAttribute<int> *nrclass = NULL;
  NRSimpleAttribute<int> *requestId = NULL;
  int retval;
  int request_number;

  nrclass = NRClassAttr.find(data);
  requestId = UserRqIdAttr.find(data);

  if (!nrclass || !requestId){
    diffPrint(DEBUG_ALWAYS, "Received a broken message !\n");
    return;
  }

  getTime(&tmv);

  switch (nrclass->getVal()){

  case NRAttribute::INTEREST_CLASS:

    request_number = requestId->getVal();

    diffPrint(DEBUG_ALWAYS, "Received new request from the user !\n");
    fprintf(ofile, "%ld:%ld - Received Request #%d\n",
	    tmv.tv_sec, tmv.tv_usec, request_number);
    fflush(NULL);

    attrs = new NRAttrVec;
    attrs->push_back(UserRqIdAttr.make(NRAttribute::EQ,
				       request_number));

    // This can go away after testing
    attrs->push_back(AppCounterAttr.make(NRAttribute::IS, send_counter));
    fprintf(ofile, "%ld:%ld - SND - N: %d - P: %d !\n",
	    tmv.tv_sec, tmv.tv_usec, node_id, send_counter);
    fflush(NULL);

    // Send the attributes to the sink
    retval = dr->send(publishHandle, attrs);
    send_counter++;

    ClearAttrs(attrs);
    delete attrs;

    break;

  default:

    diffPrint(DEBUG_ALWAYS,
	      "Received an unknown message (%d)!\n", nrclass->getVal());
    break;

  }
}

handle AcousticApp::setupUserInterest()
{
  NRAttrVec attrs;
 
  attrs.push_back(NRClassAttr.make(NRAttribute::EQ, NRAttribute::INTEREST_CLASS));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "ac"));
 
  handle h = dr->subscribe(&attrs, mr);

  ClearAttrs(&attrs);

  return h;
}

handle AcousticApp::setupPublication()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "ac"));

  handle h = dr->publish(&attrs);

  ClearAttrs(&attrs);

  return h;
}

char * AcousticApp::getSoundSample()
{
  return strdup("Boing !");
}

AcousticApp::AcousticApp(int _node_id, long _stop_time, u_int16_t port)
{
  struct timeval tmv;
  TimerType *timer;

  dr = NR::createNR(port);
  mr = new MyReceive;
  tcb = new MyTimerReceive;

  // Initialize Acoustic Application
  send_counter = 0;
  stop_time = _stop_time;
  node_id = _node_id;

  userSubscribeHandle = setupUserInterest();
  publishHandle = setupPublication();

  // Open output file
  ofile = fopen("/tmp/acoustic-src.out","w");
  if (ofile == NULL){
    diffPrint(DEBUG_ALWAYS, "Can't open log file !\n");
    exit(-1);
  }

  getTime(&tmv);
  fprintf(ofile, "%ld:%ld - Acoustic Node %d starting\n", tmv.tv_sec, tmv.tv_usec, node_id);

  // Stop this application after 'stop_time' seconds
  if (stop_time > 0){
    timer = new TimerType(STOP_TIMER);
    ((DiffusionRouting *)dr)->addTimer(stop_time * 1000, (void *) timer, tcb);
  }
}

void usage()
{
  diffPrint(DEBUG_ALWAYS, "Usage: acousticapp [-t stoptime] [-h]\n\n");
  diffPrint(DEBUG_ALWAYS, "\t-t - Schedule acoustic app to exit after stoptime seconds\n");
  diffPrint(DEBUG_ALWAYS, "\t-h - Prints this information\n");
  diffPrint(DEBUG_ALWAYS, "\n");
  exit(0);
}

void AcousticApp::run()
{
  // Do nothing
  while(1)
    sleep(100);
}

int main(int argc, char **argv)
{
  int id;
  int opt;
  long stop_time = 0;
  char *scadds;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;

  application_id = strdup(argv[0]);

#ifdef BBN_LOGGER
  InitMainLogger();
#endif

  scadds = getenv("scadds_addr");
  if (scadds == NULL){
    diffPrint(DEBUG_ALWAYS, "Error: Can't get node id !\n");
    exit(-1);
  }

  id = atoi(scadds);

  // Parse command line options
  opterr = 0;

  while (1) {
    opt = getopt(argc, argv, "ht:p:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	diffPrint(DEBUG_ALWAYS, "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 't':

      stop_time = atol(optarg);
      if (stop_time <= 0){
	diffPrint(DEBUG_ALWAYS, "Error: stop time must be > 0\n");
	exit(0);
      }
      else{
	diffPrint(DEBUG_ALWAYS,
		  "Acoustic node will stop after %ld seconds\n", stop_time); 
      }

      break;

    case 'h':

      usage();

      break;

    case '?':

      diffPrint(DEBUG_ALWAYS, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage();

      break;

    case ':':

      diffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage();

      break;

   }

    if (opt == -1)
      break;
  }

  acoustic_app = new AcousticApp(id, stop_time, diffusion_port);
  acoustic_app->run();

  return 0;
}
