// *********************************************************
//
// comApp2.cc     : Command Application for Nested Queries
// author         : Fabio Silva
//
// $Id: comApp2.cc,v 1.1 2001/11/21 19:51:25 fabio Exp $
//
// *********************************************************

#include "nestedBAE.hh"

extern char *application_id;

ComApp *com_app;

void MyReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  com_app->recv(data, my_handle);
}

void ComApp::recv(NRAttrVec *data, handle h)
{
  if (h == subscribeImagerHandle){
    ProcessImagerData(data);
    return;
  }

  diffPrint(DEBUG_ALWAYS, "Received a message for an unknown handle !\n");
}

void ComApp::ProcessImagerData(NRAttrVec *data)
{
  NRSimpleAttribute<int> *nodeIdAttr = NULL;
  NRSimpleAttribute<int> *counterAttr = NULL;
  NRSimpleAttribute<void *> *timeAttr = NULL;
  int sensor_id;
  int sensor_seq;
  long delay_seconds;
  long delay_useconds;
  float total_delay;
  struct timeval tmv;
  EventTime *sensor_event_time;

  nodeIdAttr = NodeIdAttr.find(data);
  counterAttr = AppCounterAttr.find(data);
  timeAttr = TimeAttr.find(data);

  if (!nodeIdAttr || !counterAttr || !timeAttr){
    diffPrint(DEBUG_ALWAYS, "Received a BAD message !\n");
    printAttrs(data);
    return;
  }

  getTime(&tmv);

  // Read Attributes
  sensor_seq = counterAttr->getVal();
  sensor_id = nodeIdAttr->getVal();

  // Calculate latency
  sensor_event_time = (EventTime *) timeAttr->getVal();
  delay_seconds = tmv.tv_sec;
  delay_useconds = tmv.tv_usec;

  if ((delay_seconds < sensor_event_time->seconds) ||
      ((delay_seconds == sensor_event_time->seconds) &&
       (delay_useconds < sensor_event_time->useconds))){
    // Something is wrong with the time
    diffPrint(DEBUG_IMPORTANT, "Error calculating delay !\n");
    delay_useconds = 0;
    delay_seconds = -1;
  }
  else{
    delay_seconds = delay_seconds - sensor_event_time->seconds;
    if (delay_useconds < sensor_event_time->useconds){
      delay_seconds--;
      delay_useconds = delay_useconds + 1000000;
    }
    delay_useconds = delay_useconds - sensor_event_time->useconds;
  }

  total_delay = (float) (1.0 * delay_seconds) + ((float) delay_useconds / 1000000.0);

  diffPrint(DEBUG_ALWAYS, "Received Image %d, Sensor %d !\n",
	    sensor_seq, sensor_id);
  diffPrint(DEBUG_ALWAYS, "Delay = %f\n", total_delay);

#ifdef DEBUG
  if (outfile){
    fprintf(outfile, "Node %d received image %d, sensor %d at time: %.6lf !\n",
	    node_id, sensor_seq, sensor_id, (double) tmv.tv_sec +
	    (0.000001 * tmv.tv_usec));
    fprintf(outfile, "Calculated Delay = %f\n", total_delay);
    fflush(NULL);
      }
#endif

  imager_counter++;
}

handle ComApp::setupImagerInterest()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::INTEREST_CLASS));
  attrs.push_back(TaskAttr.make(NRAttribute::EQ, "img"));

  handle h = dr->subscribe(&attrs, mr);

  ClearAttrs(&attrs);

  return h;
}

ComApp::ComApp(int _node_id, u_int16_t port)
{
  dr = NR::createNR(port);
  mr = new MyReceive;
  node_id = _node_id;
  imager_counter = 0;

  // Setup interest
  subscribeImagerHandle = setupImagerInterest();

#ifdef DEBUG
  char filename[100];

  snprintf(filename, 100, "/sensit/out/isi/isi-w-com2-%d", node_id);
  outfile = fopen(filename, "w");
#endif
  
  diffPrint(DEBUG_ALWAYS, "Command Application starting...\n");
}

void ComApp::run()
{
  while (1){
    // This thread does nothing !
    sleep (1000);
  }
}

void usage(char *s)
{
  diffPrint(DEBUG_ALWAYS, "Usage: %s [-p port] [-h]\n\n", s);
  diffPrint(DEBUG_ALWAYS, "\t-p - Specifies the diffusion port on the node\n");
  diffPrint(DEBUG_ALWAYS, "\t-h - Prints this information\n");
  diffPrint(DEBUG_ALWAYS, "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  int id;
  int opt;
  char *scadds;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;

  application_id = strdup(argv[0]);

#ifdef BBN_LOGGER
  InitMainLogger();
#endif

  scadds = getenv("scadds_addr");
  if (scadds == NULL){
    diffPrint(DEBUG_ALWAYS, "Error: Can't get node id !\n");
    exit(-1);
  }

  id = atoi(scadds);

  // Parse command line options
  opterr = 0;

  while (1) {
    opt = getopt(argc, argv, "hp:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	diffPrint(DEBUG_ALWAYS,
		  "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage(argv[0]);

      break;

    case '?':

      diffPrint(DEBUG_ALWAYS, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage(argv[0]);

      break;

    case ':':

      diffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage(argv[0]);

      break;

   }

    if (opt == -1)
      break;
  }

  com_app = new ComApp(id, diffusion_port);
  com_app->run();

  return 0;
}
