// *********************************************************
//
// imgApp2.cc      : Imager App for nested queries
// author          : Fabio Silva
//
// $Id: imgApp2.cc,v 1.1 2001/11/21 19:51:26 fabio Exp $
//
// *********************************************************

#include "nestedBAE.hh"
#include <unistd.h>

extern char *application_id;

ImagerApp *imager_app;

void MyReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  imager_app->recv(data, my_handle);
}

void ImagerApp::recv(NRAttrVec *data, handle h)
{
  if (h == userSubscribeHandle){
    ProcessUserRequest(data);
    return;
  }

  if (h == lightInfoHandle){
    ProcessSensorInfo(data);
    return;
  }

  diffPrint(DEBUG_ALWAYS, "Received a message for an unknown handle !\n");
}

void ImagerApp::ProcessUserRequest(NRAttrVec *data)
{
  NRSimpleAttribute<int> *nrclass = NULL;
  struct timeval tv;
  int retval;

  nrclass = NRClassAttr.find(data);

  if (!nrclass){
    diffPrint(DEBUG_ALWAYS, "Received a broken message !\n");
    return;
  }

  getTime(&tv);

  switch (nrclass->getVal()){

  case NRAttribute::INTEREST_CLASS:

    diffPrint(DEBUG_ALWAYS,
	      "Received an Interest message from the user node !\n");

    // Start Interest for light information if not tasked
    tasked++;

    if (tasked == 1){

      diffPrint(DEBUG_ALWAYS, "Imager App tasked by user !\n");

#ifdef DEBUG
      if (outfile){
	fprintf(outfile, "Imager %d tasked by user at time: %.6lf !\n",
		node_id, (double) tv.tv_sec + (0.000001 * tv.tv_usec));
	fflush(NULL);
      }
#endif

      // Sleep for a while to avoid synchronization
      sleep(5);

      // Setup light interest
      lightInfoHandle = setupLightInterest();
    }

    break;

  case NRAttribute::DISINTEREST_CLASS:

    diffPrint(DEBUG_ALWAYS,
	      "Received a Disinterest message from the user node !\n");

    // Stop collecting light if no tasks left
    tasked--;

    if (tasked == 0){

      diffPrint(DEBUG_ALWAYS, "Imager App untasked by user !\n");

#ifdef DEBUG
      if (outfile){
	fprintf(outfile, "Imager %d untasked by user at time: %.6lf !\n",
		node_id, (double) tv.tv_sec + (0.000001 * tv.tv_usec));
	fflush(NULL);
      }
#endif

      // Unsubscribe for light information
      retval = dr->unsubscribe(lightInfoHandle);

      if (retval != OK)
	diffPrint(DEBUG_ALWAYS,
		  "Failed to unsubscribe to light information !\n");
    }

    break;

  default:

    diffPrint(DEBUG_ALWAYS,
	      "Received an unknown message (%d)!\n", nrclass->getVal());

    break;

  }
}

void ImagerApp::ProcessSensorInfo(NRAttrVec *data)
{
  int sensor_id;
  int sensor_seq;
  NRSimpleAttribute<int> *counterAttr;
  NRSimpleAttribute<void *> *timeAttr;
  NRSimpleAttribute<int> *nodeIdAttr;
  struct timeval tv;
  NRAttrVec *attrs;
  int retval;

  // Return if not currently tasked
  if (tasked == 0)
    return;

  // Locate Info
  counterAttr = AppCounterAttr.find(data);
  timeAttr = TimeAttr.find(data);
  nodeIdAttr = NodeIdAttr.find(data);

  if (!counterAttr || !timeAttr || !nodeIdAttr){
    diffPrint(DEBUG_ALWAYS,
	      "Message received from light sensor is broken !\n");
    return;
  }

  sensor_seq = counterAttr->getVal();
  sensor_id = nodeIdAttr->getVal();

  receive_counter++;

  getTime(&tv);

  // Output message received from light sensor
  diffPrint(DEBUG_ALWAYS, "Received CPA %d from node %d !\n",
	    sensor_seq, sensor_id);

  // Prepare attributes to send
  attrs = new NRAttrVec;
  attrs->push_back(TimeAttr.make(NRAttribute::IS, timeAttr->getVal(), sizeof(EventTime)));
  attrs->push_back(AppCounterAttr.make(NRAttribute::IS, send_counter));
  attrs->push_back(NodeIdAttr.make(NRAttribute::IS, sensor_id));

  diffPrint(DEBUG_ALWAYS, "Sending to user node, sensor id: %d, seq: %d, imager id: %d !\n", sensor_id, sensor_seq, send_counter);

#ifdef DEBUG
  if (outfile){
    fprintf(outfile, "Received CPA %d from node %d, time: %.6lf !\n",
	    sensor_seq, sensor_id, (double) tv.tv_sec +
	    (0.000001 * tv.tv_usec));
    fprintf(outfile, "Sending Image %d !\n", send_counter);
    fflush(NULL);
  }
#endif

  // Send the attributes to the User Node
  retval = dr->send(publishHandle, attrs);
  send_counter++;

  ClearAttrs(attrs);

  delete attrs;
}

handle ImagerApp::setupLightInterest()
{
  NRAttrVec lightAttrs;

  lightAttrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::INTEREST_CLASS));
  lightAttrs.push_back(TaskAttr.make(NRAttribute::EQ, "pir"));

  handle h = dr->subscribe(&lightAttrs, mr);

  ClearAttrs(&lightAttrs);

  return h;
}

handle ImagerApp::setupUserInterest()
{
  NRAttrVec attrs;
 
  attrs.push_back(NRClassAttr.make(NRAttribute::NE, NRAttribute::DATA_CLASS));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "img"));
 
  handle h = dr->subscribe(&attrs, mr);

  ClearAttrs(&attrs);

  return h;
}

handle ImagerApp::setupPublication()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "img"));

  handle h = dr->publish(&attrs);

  ClearAttrs(&attrs);

  return h;
}

char * ImagerApp::getImage()
{
  return strdup("Look at this tank !");
}

ImagerApp::ImagerApp(int _node_id, u_int16_t port)
{
  dr = NR::createNR(port);
  mr = new MyReceive;

  // Initialize Acoustic Application
  tasked = 0;
  send_counter = 0;
  receive_counter = 0;
  node_id = _node_id;

  userSubscribeHandle = setupUserInterest();
  publishHandle = setupPublication();

#ifdef DEBUG
  char filename[100];

  snprintf(filename, 100, "/sensit/out/isi/isi-w-img2-%d", node_id);
  outfile = fopen(filename, "w");
#endif

  diffPrint(DEBUG_ALWAYS, "Imager App starting...\n");
}

void usage(char *s)
{
  diffPrint(DEBUG_ALWAYS, "Usage: %s [-p port] [-h]\n\n");
  diffPrint(DEBUG_ALWAYS, "\t-p - Specifies the diffusion port on the node\n");
  diffPrint(DEBUG_ALWAYS, "\t-h - Prints this information\n");
  diffPrint(DEBUG_ALWAYS, "\n");
  exit(0);
}

void ImagerApp::run()
{
  // Do nothing
  while(1)
    sleep(100);
}

int main(int argc, char **argv)
{
  int id;
  int opt;
  char *scadds;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;

  application_id = strdup(argv[0]);

#ifdef BBN_LOGGER
  InitMainLogger();
#endif

  scadds = getenv("scadds_addr");
  if (scadds == NULL){
    diffPrint(DEBUG_ALWAYS, "Error: Can't get node id !\n");
    exit(-1);
  }

  id = atoi(scadds);

  // Parse command line options
  opterr = 0;

  while (1) {
    opt = getopt(argc, argv, "hp:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	diffPrint(DEBUG_ALWAYS,
		  "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage(argv[0]);

      break;

    case '?':

      diffPrint(DEBUG_ALWAYS, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage(argv[0]);

      break;

    case ':':

      diffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage(argv[0]);

      break;

   }

    if (opt == -1)
      break;
  }

  imager_app = new ImagerApp(id, diffusion_port);
  imager_app->run();

  return 0;
}
