// *********************************************************
//
// userApp2.cc    : User Application for Nested Queries
// author         : Fabio Silva
//
// $Id: userApp2.cc,v 1.1 2001/11/21 19:51:28 fabio Exp $
//
// *********************************************************

#include "nested.hh"

extern char *application_id;

UserApp *user_app;

int MyTimerReceive::recv(handle hdl, void *p)
{
  return user_app->ProcessTimers(hdl, p);
}

void MyTimerReceive::del(void *p)
{
  // Nothing to do !
}

void MyReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  user_app->recv(data, my_handle);
}

int UserApp::ProcessTimers(handle hdl, void *p)
{
  TimerType *timer;

  timer = (TimerType *) p;

  switch (timer->which_timer){

  case STOP_TIMER:

    timetoStop();
    exit(0);

    break;
  }
  return 0;
}

void UserApp::timetoStop()
{
  struct timeval tv;

  getTime(&tv);
  fprintf(ofile, "%ld:%ld - Sink %d stopping\n", tv.tv_sec, tv.tv_usec, node_id);
  fprintf(ofile, "Total messages received : %d\n", acoustic_counter);
  fflush(NULL);
  fclose(ofile);
}

void UserApp::recv(NRAttrVec *data, handle h)
{
  if (h == subscribeAcousticHandle){
    ProcessAcousticData(data);
    return;
  }

  diffPrint(DEBUG_ALWAYS, "Received a message for an unknown handle !\n");
}

void UserApp::ProcessAcousticData(NRAttrVec *data)
{
  NRSimpleAttribute<void *> *lightAttr = NULL;
  NRSimpleAttribute<int> *counterAttr = NULL;
  NRSimpleAttribute<void *> *timeAttr = NULL;
  Light_Node_State *light_state;
  int sensor_id;
  int sensor_state;
  int sensor_seq;
  long delay_seconds;
  long delay_useconds;
  float total_delay;
  struct timeval tmv;
  EventTime *sensor_event_time;

  lightAttr = LightStateAttr.find(data);
  counterAttr = AppCounterAttr.find(data);
  timeAttr = TimeAttr.find(data);

  if (!lightAttr || !counterAttr || !timeAttr){
    diffPrint(DEBUG_ALWAYS, "Received a BAD message !\n");
    printAttrs(data);
    return;
  }

  getTime(&tmv);

  // Read Attributes
  light_state = (Light_Node_State *) lightAttr->getVal();
  sensor_seq = counterAttr->getVal();

  sensor_id = light_state->node_id;
  sensor_state = light_state->light_state;

  // Calculate latency
  sensor_event_time = (EventTime *) timeAttr->getVal();
  delay_seconds = tmv.tv_sec;
  delay_useconds = tmv.tv_usec;

  if ((delay_seconds < sensor_event_time->seconds) ||
      ((delay_seconds == sensor_event_time->seconds) &&
       (delay_useconds < sensor_event_time->useconds))){
    // Something is wrong with the time
    fprintf(ofile, "Error calculating delay !\n");
    fflush(NULL);
  }
  else{
    delay_seconds = delay_seconds - sensor_event_time->seconds;
    if (delay_useconds < sensor_event_time->useconds){
      delay_seconds--;
      delay_useconds = delay_useconds + 1000000;
    }
    delay_useconds = delay_useconds - sensor_event_time->useconds;
  }

  total_delay = (float) (1.0 * delay_seconds) + ((float) delay_useconds / 1000000.0);

  fprintf(ofile, "%ld:%ld - RCVD - N: %d - P: %d - L: ", tmv.tv_sec, tmv.tv_usec, sensor_id, sensor_seq);
  if (sensor_state == LIGHT_IS_ON){
    fprintf(ofile, "ON !\n");
  }
  else{
    fprintf(ofile, "OFF !\n");
  }
  fprintf(ofile, "Delay = %f\n", total_delay);
  fflush(NULL);

  acoustic_counter++;
}

handle UserApp::setupAcousticInterest()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::INTEREST_CLASS));
  attrs.push_back(TaskAttr.make(NRAttribute::EQ, "ac"));

  handle h = dr->subscribe(&attrs, mr);

  ClearAttrs(&attrs);

  return h;
}

UserApp::UserApp(int _node_id, long _stop_time, u_int16_t port)
{
  struct timeval tmv;
  TimerType *timer;

  dr = NR::createNR(port);
  mr = new MyReceive;
  tcb = new MyTimerReceive;
  node_id = _node_id;
  stop_time = _stop_time;
  acoustic_counter = 0;

  // Setup interest
  subscribeAcousticHandle = setupAcousticInterest();
  
  // Open output file
  ofile = fopen("/tmp/user-sink.out","w");
  if (ofile == NULL){
    diffPrint(DEBUG_ALWAYS, "Can't open log file !\n");
    exit(-1);
  }

  getTime(&tmv);
  fprintf(ofile, "%ld:%ld - User Application %d starting\n",
	  tmv.tv_sec, tmv.tv_usec, node_id);

  // Stop this application after 'stop_time' seconds
  if (stop_time > 0){
    timer = new TimerType(STOP_TIMER);
    ((DiffusionRouting *)dr)->addTimer(stop_time * 1000, (void *) timer, tcb);
  }
}

void UserApp::run()
{
  while (1){
    // This thread does nothing !
    sleep (1000);
  }
}

void usage()
{
  diffPrint(DEBUG_ALWAYS, "Usage: userapp [-t stoptime] [-h]\n\n");
  diffPrint(DEBUG_ALWAYS, "\t-t - Schedule user app to exit after stoptime seconds\n");
  diffPrint(DEBUG_ALWAYS, "\t-h - Prints this information\n");
  diffPrint(DEBUG_ALWAYS, "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  int id;
  int opt;
  long stop_time = 0;
  char *scadds;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;

  application_id = strdup(argv[0]);

#ifdef BBN_LOGGER
  InitMainLogger();
#endif

  scadds = getenv("scadds_addr");
  if (scadds == NULL){
    diffPrint(DEBUG_ALWAYS, "Error: Can't get node id !\n");
    exit(-1);
  }

  id = atoi(scadds);

  // Parse command line options
  opterr = 0;

  while (1) {
    opt = getopt(argc, argv, "ht:p:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	diffPrint(DEBUG_ALWAYS,
		  "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 't':

      stop_time = atol(optarg);
      if (stop_time <= 0){
	diffPrint(DEBUG_ALWAYS, "Error: stop time must be > 0\n");
	exit(0);
      }
      else{
	diffPrint(DEBUG_ALWAYS,
		  "User app will stop after %ld seconds\n", stop_time); 
      }

      break;

    case 'h':

      usage();

      break;

    case '?':

      diffPrint(DEBUG_ALWAYS, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage();

      break;

    case ':':

      diffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage();

      break;

   }

    if (opt == -1)
      break;
  }

  user_app = new UserApp(id, stop_time, diffusion_port);
  user_app->run();

  return 0;
}
