// *********************************************************
//
// sensorApp.cc   : Sensor App for nested queries
// author         : Fabio Silva
//
// $Id: sensorApp.cc,v 1.1 2001/11/21 19:51:29 fabio Exp $
//
// *********************************************************

#include "sk.hh"
#include <unistd.h>

extern char *application_id;

SensorApp *sensor_app = NULL;

void rp(REP_EVENT ev);

void MyReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  sensor_app->recv(data, my_handle);
}

void SensorApp::recv(NRAttrVec *data, handle h)
{
  NRSimpleAttribute<int> *nrclass = NULL;

  if (h == subscribeHandle){
    nrclass = NRClassAttr.find(data);

    if (nrclass){
      switch (nrclass->getVal()){

      case NRAttribute::INTEREST_CLASS:

	diffPrint(DEBUG_ALWAYS, "Tasked by user !\n");
	tasked++;

	break;

      case NRAttribute::DISINTEREST_CLASS:

	diffPrint(DEBUG_ALWAYS, "Untasked !\n");
	tasked--;

	break;

      default:

	diffPrint(DEBUG_ALWAYS, "Received unknown message !\n");

	break;
      }
    }
    else{
      diffPrint(DEBUG_ALWAYS, "Received a broken message !\n");
    }
  }
  else{
    diffPrint(DEBUG_ALWAYS, "Received a message for an unknown handle !\n");
  }
}

void SensorApp::sensorCallback(REP_EVENT ev){
  NRSimpleAttribute<void *> *cpaAttr;
  NRAttrVec data_attr;
  CpaEvent cpa_event;
  int retval;
  HighLevel          highlevel;
  REP_RECORD_HEADER  rechdr;

  RepGetRecordHeader(ev.repid, ev.pos, &rechdr);
      
  // Process sensor callback
  switch(rechdr.structid){

  case REP_HIGH_LEVEL:
    RepGetHighLevelEvent(ev.repid, ev.pos, &highlevel);

    // Log CPA event
    diffPrint(DEBUG_IMPORTANT, "Got a CPA, Sensor = %d !\n",
	      highlevel.sensorid);

    if (tasked){
      // Got a PIR detection. Send it out
      events_counter++;

      cpa_event.node_id = node_id;
      cpa_event.sensor_id = highlevel.sensorid;
      cpa_event.channel_id = highlevel.channelid;
      cpa_event.direction = highlevel.direction;
      cpa_event.confidence = highlevel.confidence;
      cpa_event.cpa_time = highlevel.timestamp.tv_sec +
	(0.000001 * highlevel.timestamp.tv_usec);

      cpaAttr = CpaAttr.make(NRAttribute::IS, (void *) &cpa_event,
			     sizeof(CpaEvent));

      data_attr.push_back(cpaAttr);

      retval = dr->send(publishHandle, &data_attr);

      diffPrint(DEBUG_ALWAYS, "Sending CPA Event %d !\n", events_counter);

#ifdef DEBUG
      if (outfile){
	fprintf(outfile, "Node %d sent event %d, time: %.6lf !\n",
		node_id, events_counter, cpa_event.cpa_time);
	fflush(NULL);
      }
#endif

      ClearAttrs(&data_attr);
    }

    break;
  }
}

handle SensorApp::setupPublication()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "cpa"));

  handle h = dr->publish(&attrs);

  ClearAttrs(&attrs);

  return h;
}

handle SensorApp::setupSubscription()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::NE, NRAttribute::DATA_CLASS));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(TaskAttr.make(NRAttribute::IS, "cpa"));

  handle h = dr->subscribe(&attrs, mr);

  ClearAttrs(&attrs);

  return h;
}

SensorApp::SensorApp(int _node_id, u_int16_t port)
{
  // Initialize
  node_id = _node_id;
  tasked = 0;
  events_counter = 0;

  // Create classes and set up publication
  dr = NR::createNR(port);
  mr = new MyReceive;
  publishHandle = setupPublication();
  subscribeHandle = setupSubscription();

#ifdef DEBUG
  char filename[100];

  snprintf(filename, 100, "/sensit/out/isi/isi-w-sensit-%d", node_id);
  outfile = fopen(filename, "w");
#endif

  diffPrint(DEBUG_ALWAYS, "Sensor App starting...\n");
}

void SensorApp::run()
{
  // Main thread does nothing !
  while(1){
    sleep(100);
  }
}

void usage(char *s)
{
  diffPrint(DEBUG_ALWAYS, "Usage: %s [-p port] [-h]\n\n", s);
  diffPrint(DEBUG_ALWAYS, "\t-p - Specifies the diffusion port on the node\n");
  diffPrint(DEBUG_ALWAYS, "\t-h - Prints this information\n");
  diffPrint(DEBUG_ALWAYS, "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  int id;
  int opt;
  char *scadds;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;
  CSubscriptionManager sm;
  REP_RESULT           ret;
  CSubList             sublist;

  application_id = strdup(argv[0]);

#ifdef BBN_LOGGER
  InitMainLogger();
#endif

  scadds = getenv("scadds_addr");
  if (scadds == NULL){
    diffPrint(DEBUG_ALWAYS, "Error: Can't get node id !\n");
    exit(-1);
  }

  id = atoi(scadds);

  // Parse command line options
  opterr = 0;

  while (1) {
    opt = getopt(argc, argv, "hp:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	diffPrint(DEBUG_ALWAYS,
		  "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage(argv[0]);

      break;

    case '?':

      diffPrint(DEBUG_ALWAYS, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage(argv[0]);

      break;

    case ':':

      diffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage(argv[0]);

      break;

   }

    if (opt == -1)
      break;
  }

  sensor_app = new SensorApp(id, diffusion_port);

  // Initialize communication with BAE repository
  InitRepository();
  CreateRepository();

  ret = sm.Init("ISI-Client", rp);
  if(ret != REP_OK){
    diffPrint(DEBUG_ALWAYS, "Init error\n");
    exit(-1);
  }

  sublist.Subscribe(HIGH_LEVEL_REP);

  ret = sm.SetSubscription(&sublist);
  if(ret != REP_OK){
    diffPrint(DEBUG_ALWAYS, "Set subscription error\n");
    exit(-1);
  }

  SetSubscriptionManager(&sm); 

  sensor_app->run();
}

void rp(REP_EVENT ev){
  sensor_app->sensorCallback(ev);
}
